﻿using System;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Windows.Forms;

namespace GBAFE_Tilesets
{
    public partial class TilesetPaletteForm : Form
    {
        const int SWATCH_SIZE = 16;

        private int TilesetIndex = 0;
        private Color[] Palette;
        private bool SelectingIndex = false;

        internal Color[] palette
        {
            get { return Palette; }
            set
            {
                if (value == null)
                {
                    Palette = null;
                    TilesetIndex = 0;
                    enable_palette_controls(false);
                }
                else
                {
                    Palette = new Color[value.Length];
                    Array.Copy(value, Palette, Palette.Length);
                    TilesetIndex = Math.Max(0, Math.Min(TilesetIndex, Palette.Length - 1));
                    colorSelector1.active_color = Palette[TilesetIndex];
                    enable_palette_controls(true);
                }
                PalettePictureBox.Invalidate();
            }
        }

        public TilesetPaletteForm(Form owner)
        {
            Owner = owner;
            InitializeComponent();

            PalettePictureBox.MouseDown += PalettePictureBox_MouseDown;
            PalettePictureBox.MouseMove += PalettePictureBox_MouseMove;
            PalettePictureBox.MouseUp += PalettePictureBox_MouseUp;
            PalettePictureBox.Paint += PalettePictureBox_Paint;
            colorSelector1.ColorEdited += colorSelector1_ColorEdited;
        }

        private void select_index(Point loc)
        {
            if (loc.X >= 0 && loc.X < PalettePictureBox.Width && loc.Y >= 0 && loc.Y < PalettePictureBox.Height)
            {
                TilesetIndex = (loc.Y / 16) * 16 + (loc.X / 16);
                colorSelector1.active_color = Palette[TilesetIndex];
                enable_palette_controls(true);
                PalettePictureBox.Invalidate();
            }
        }

        private void enable_palette_controls(bool enabled)
        {
            colorSelector1.Enabled = enabled;
            SavePaletteButton.Enabled = enabled;
            LoadPaletteButton.Enabled = enabled;
            CopyPaletteHexButton.Enabled = enabled;
            CopyPaletteImageButton.Enabled = enabled;
            PastePaletteImageButton.Enabled = enabled;
        }

        internal event EventHandler PaletteChanged;

        #region Events
        void PalettePictureBox_Paint(object sender, PaintEventArgs e)
        {
            if (Palette != null)
            {
                for (int i = 0; i < Palette.Length && i < 160; i++)
                {
                    using (Brush b = new SolidBrush(Palette[i]))
                        e.Graphics.FillRectangle(b, new Rectangle((i % 16) * SWATCH_SIZE, (i / 16) * SWATCH_SIZE, SWATCH_SIZE, SWATCH_SIZE));
                }
                if (TilesetIndex >= 0)
                {
                    using (Pen p = new Pen(Color.Black, 4))
                        e.Graphics.DrawRectangle(p, new Rectangle((TilesetIndex % 16) * SWATCH_SIZE + 2, (TilesetIndex / 16) * SWATCH_SIZE + 2, SWATCH_SIZE - 4, SWATCH_SIZE - 4));
                    using (Pen p = new Pen(Color.White, 2))
                        e.Graphics.DrawRectangle(p, new Rectangle((TilesetIndex % 16) * SWATCH_SIZE + 2, (TilesetIndex / 16) * SWATCH_SIZE + 2, SWATCH_SIZE - 4, SWATCH_SIZE - 4));
                }
            }
        }

        void PalettePictureBox_MouseDown(object sender, MouseEventArgs e)
        {
            if (Palette != null && e.Button == MouseButtons.Left)
            {
                SelectingIndex = true;
                select_index(e.Location);
            }
        }
        void PalettePictureBox_MouseMove(object sender, MouseEventArgs e)
        {
            if (SelectingIndex)
            {
                select_index(e.Location);
            }
        }
        void PalettePictureBox_MouseUp(object sender, MouseEventArgs e)
        {
            SelectingIndex = false;
        }

        void colorSelector1_ColorEdited(object sender, EventArgs e)
        {
            Palette[TilesetIndex] = colorSelector1.active_color;
            PalettePictureBox.Invalidate();
            PaletteChanged(this, new EventArgs());
        }

        private void SavePaletteButton_Click(object sender, EventArgs e)
        {
            if (PaletteSaveDialog.ShowDialog() == DialogResult.OK)
            {
                using (FileStream stream = new FileStream(PaletteSaveDialog.FileName, FileMode.Create))
                using (BinaryWriter writer = new BinaryWriter(stream))
                {
                    byte[] hex_palette = YetiGraphics.color_array_to_hex(Palette);
                    writer.Write(hex_palette);
                }
            }
        }
        private void LoadPaletteButton_Click(object sender, EventArgs e)
        {
            if(PaletteOpenDialog.ShowDialog() == DialogResult.OK)
            {
                using (FileStream stream = new FileStream(PaletteOpenDialog.FileName, FileMode.Open))
                using (BinaryReader reader = new BinaryReader(stream))
                {
                    if (reader.BaseStream.Length >= Palette.Length * 2)
                    {
                        byte[] hex_palette = reader.ReadBytes(Palette.Length * 2);
                        Color[] palette = YetiGraphics.hex_to_color_array(hex_palette);
                        this.palette = palette;
                        PaletteChanged(this, new EventArgs());
                    }
                }
            }
        }

        private void CopyPaletteHexButton_Click(object sender, EventArgs e)
        {
            byte[] hex_palette = YetiGraphics.color_array_to_hex(Palette);
            string result = hex_palette.Aggregate("", (str, x) => str + x.ToString("X2"));
            Clipboard.SetText(result);
        }

        private void CopyPaletteImageButton_Click(object sender, EventArgs e)
        {
            using (Bitmap bmp = new Bitmap(16, (int)Math.Ceiling(Palette.Length / 16f)))
            {
                for (int i = 0; i < Palette.Length; i++)
                    bmp.SetPixel(i % bmp.Width, i / bmp.Width, Palette[i]);
                Clipboard.SetImage(bmp);
            }
        }

        private void PastePaletteImageButton_Click(object sender, EventArgs e)
        {
            if (Clipboard.ContainsImage())
                using (Image bmp = Clipboard.GetImage())
                    if (bmp is Bitmap && bmp.Width * bmp.Height >= Palette.Length)
                    {
                        Color[] palette = new Color[Palette.Length];
                        for (int i = 0; i < palette.Length; i++)
                            palette[i] = YetiGraphics.convert_color_bit_depth(
                                Color.FromArgb(255, (bmp as Bitmap).GetPixel(i % (bmp as Bitmap).Width, i / (bmp as Bitmap).Width)),
                                Color_Bit_Depth.R5G5B5, Color_Bit_Depth.R8G8B8);
                        this.palette = palette;
                        PaletteChanged(this, new EventArgs());
                    }
        }
        #endregion
    }
}
