﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Imaging;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Windows.Forms;
using Nintenlord.ROMHacking.GBA;
using Nintenlord.ROMHacking.GBA.Graphics;
using Nintenlord.Forms.Utility;

namespace GBAFE_Tilesets
{
    enum FEVersion { FE6, FE7, FE8, Unknown }
    public partial class TilesetViewerForm : Form
    {
        FEVersion Version;
        GBAROM FERom = null;
        List<TilesetData> Tilesets;
        Bitmap RawTilesetImg, TilesetImg;
        private TilesetPaletteForm PaletteForm;
        FEChapterFormatInfo Chapter_Info;

        #region Accessors
        private bool tileset_exists { get { return Tilesets != null && TilesetListBox.SelectedIndex >= 0 && TilesetListBox.SelectedIndex < Tilesets.Count; } }
        private TilesetData tileset { get { return Tilesets[TilesetListBox.SelectedIndex]; } }
        private Color[] tileset_palette { get { return this.tileset.Palettes[PaletteListBox.SelectedIndex].Item2; } }

        private bool main_chapter_entry_valid
        {
            get
            {
                int result;
                string s = MainChapterTextBox.Text.Length >= 2 && MainChapterTextBox.Text.Substring(0, 2) == "0x" ?
                    MainChapterTextBox.Text.Substring(2, MainChapterTextBox.Text.Length - 2) : MainChapterTextBox.Text;
                return int.TryParse(s, NumberStyles.AllowHexSpecifier, CultureInfo.InvariantCulture, out result);
            }
        }
        private int main_chapter_value
        {
            get
            {
                int result;
                string s = MainChapterTextBox.Text.Length >= 2 && MainChapterTextBox.Text.Substring(0, 2) == "0x" ?
                    MainChapterTextBox.Text.Substring(2, MainChapterTextBox.Text.Length - 2) : MainChapterTextBox.Text;
                if (!int.TryParse(s, NumberStyles.AllowHexSpecifier, CultureInfo.InvariantCulture, out result))
                    return -1;
                return result;
            }
        }

        private bool chapter_data_entry_valid
        {
            get
            {
                int result;
                string s = ChapterDataTextBox.Text.Length >= 2 && ChapterDataTextBox.Text.Substring(0, 2) == "0x" ?
                    ChapterDataTextBox.Text.Substring(2, ChapterDataTextBox.Text.Length - 2) : ChapterDataTextBox.Text;
                return int.TryParse(s, NumberStyles.AllowHexSpecifier, CultureInfo.InvariantCulture, out result);
            }
        }
        private int chapter_data_value
        {
            get
            {
                int result;
                string s = ChapterDataTextBox.Text.Length >= 2 && ChapterDataTextBox.Text.Substring(0, 2) == "0x" ?
                    ChapterDataTextBox.Text.Substring(2, ChapterDataTextBox.Text.Length - 2) : ChapterDataTextBox.Text;
                if (!int.TryParse(s, NumberStyles.AllowHexSpecifier, CultureInfo.InvariantCulture, out result))
                    return -1;
                return result;
            }
        }

        private FEChapterFormatInfo chapter_info { get { return Chapter_Info; } }
        #endregion

        public TilesetViewerForm()
        {
            InitializeComponent();

            PaletteForm = new TilesetPaletteForm(this);

            VersionComboBox.SelectedIndex = (int)FEVersion.FE7;
            Shown += TilesetViewerForm_Shown;
            RawTilesetPictureBox.Paint += RawTilesetPictureBox_Paint;
            TilesetPictureBox.Paint += TilesetPictureBox_Paint;
            PaletteForm.PaletteChanged += PaletteForm_PaletteChanged;
            refresh_chapter_info();
        }

        private void load_game(string filename)
        {
            if (!main_chapter_entry_valid || !chapter_data_entry_valid)
                return;
            refresh_chapter_info();
            var chapter_info = this.chapter_info;
            GBAROM rom = new GBAROM();
            rom.OpenROM(filename);
            var unique_tilesets = TilesetProcessing.get_unique_tilesets(rom, Version, chapter_info);
            if (unique_tilesets == null || unique_tilesets.Count <= 0)
            {
                rom.CloseROM();
                MessageBox.Show("Could not find tileset data in the rom file.\nConfirm the game setting is correct.", "Error Loading Rom", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            else
            {
                if (RawTilesetImg != null)
                {
                    RawTilesetImg.Dispose();
                    RawTilesetImg = null;
                    TilesetImg.Dispose();
                    TilesetImg = null;
                }
                if (FERom != null)
                {
                    FERom.CloseROM();
                    FERom = null;
                }
                FERom = rom;

                var tilesets_without_palettes = new HashSet<Tuple<byte, byte, byte>>(unique_tilesets
                    .Select(x => new Tuple<byte, byte, byte>(x.Item1, x.Item2, x.Item4)));
                var tileset_palettes = tilesets_without_palettes
                    .ToDictionary(p => p, p =>
                        new HashSet<byte>(unique_tilesets
                            .Where(x => p.Item1 == x.Item1 && p.Item2 == x.Item2 && p.Item3 == x.Item4)
                            .Select(x => x.Item3)));
                Tilesets = tileset_palettes
                    .Select(p => new TilesetData(p.Key,
                        TilesetProcessing.load_tileset_graphic_data(FERom, chapter_info, p.Key),
                        p.Value.Select(x => new Tuple<byte, byte[]>(x, TilesetProcessing.load_tileset_palette(FERom, chapter_info, x))).ToArray(),
                        TilesetProcessing.load_tileset_data(FERom, chapter_info, p.Key))).ToList();

                refresh_tilesets();
                RawTilesetPictureBox.Invalidate();
                TilesetPictureBox.Invalidate();
            }
        }

        private void refresh_tilesets()
        {
            TilesetListBox.Items.Clear();
            AddPaletteButton.Enabled = Tilesets.Count > 0;
            RemovePaletteButton.Enabled = false;
            foreach (var tileset in Tilesets)
                TilesetListBox.Items.Add(string.Format("Tileset {2}; Graphics {0},{1}; {3} entries",
                    tileset.Data.DataOffsets.Item1.ToString("X2"),
                    tileset.Data.DataOffsets.Item2.ToString("X2"),
                    tileset.Data.DataOffsets.Item3.ToString("X2"),
                    tileset.Palettes.Count));
            refresh_palettes();
        }

        private void refresh_palettes()
        {
            PaletteListBox.Items.Clear();
            if (tileset_exists)
                foreach (var palette in this.tileset.Palettes)
                    PaletteListBox.Items.Add(string.Format("Pal {0}",
                        palette.Item1.ToString("X2")));
        }

        private void refresh_chapter_info()
        {
            Chapter_Info = Version != FEVersion.Unknown ? TilesetProcessing.DEFAULT_CHAPTER_DATA[Version] :
                new FEChapterFormatInfo(main_chapter_value, chapter_data_value, (int)ChapterLengthSpinner.Value, (int)ChapterCountSpinner.Value);
        }

        #region Events
        private void RawTilesetUnusedCheckBox_CheckedChanged(object sender, EventArgs e)
        {
            RawTilesetPictureBox.Invalidate();
        }

        private void FogCheckBox_CheckedChanged(object sender, EventArgs e)
        {
            //change_tileset();
            refresh_palette();
        }

        private void TilesetListBox_SelectedIndexChanged(object sender, EventArgs e)
        {
            refresh_palettes();
            PaletteListBox.SelectedIndex = 0;
        }
        private void PaletteListBox_SelectedIndexChanged(object sender, EventArgs e)
        {
            change_tileset();
        }

        private void change_tileset()
        {
            if (tileset_exists && PaletteListBox.SelectedIndex >= 0)
            {
                PaletteForm.palette = tileset_palette;
                refresh_tileset_graphic();
                RemovePaletteButton.Enabled = this.tileset.Palettes.Count > 1;
            }
            else
            {
                RemovePaletteButton.Enabled = false;
                PaletteForm.palette = null;
            }
        }
        private void refresh_tileset_graphic()
        {
            if (RawTilesetImg != null)
            {
                RawTilesetImg.Dispose();
                RawTilesetImg = null;
                TilesetImg.Dispose();
                TilesetImg = null;
            }
            var chapter_info = this.chapter_info;
            TilesetImg = TilesetProcessing.load_tileset_graphic(FERom, chapter_info, this.tileset, this.tileset_palette);
            if (TilesetImg != null)
                RawTilesetImg = TilesetProcessing.load_raw_tileset_graphic(FERom, chapter_info, this.tileset, this.tileset_palette);
            refresh_palette();
        }
        private void refresh_palette()
        {
            if (RawTilesetImg != null)
            {
                TilesetProcessing.set_tileset_palette(RawTilesetImg, this.tileset_palette, FogCheckBox.Checked);
                TilesetProcessing.set_tileset_palette(TilesetImg, this.tileset_palette, FogCheckBox.Checked);
            }
            RawTilesetPictureBox.Invalidate();
            TilesetPictureBox.Invalidate();
        }

        private void AddPaletteButton_Click(object sender, EventArgs e)
        {
            if (tileset_exists)
            {
                int tileset_index = TilesetListBox.SelectedIndex;
                int index = Math.Max(0, PaletteListBox.SelectedIndex);
                tileset.copy_palette(index);
                refresh_tilesets();
                TilesetListBox.SelectedIndex = tileset_index;
                PaletteListBox.SelectedIndex = tileset.Palettes.Count - 1;
            }
        }
        private void RemovePaletteButton_Click(object sender, EventArgs e)
        {
            if (tileset_exists && tileset.Palettes.Count > 1 && PaletteListBox.SelectedIndex >= 0 && PaletteListBox.SelectedIndex < tileset.Palettes.Count)
            {
                int tileset_index = TilesetListBox.SelectedIndex;
                int index = PaletteListBox.SelectedIndex;
                tileset.remove_palette(index);
                refresh_tilesets();
                TilesetListBox.SelectedIndex = tileset_index;
                PaletteListBox.SelectedIndex = Math.Min(index, tileset.Palettes.Count - 1);
            }
        }

        private void VersionComboBox_SelectedIndexChanged(object sender, EventArgs e)
        {
            Version = (FEVersion)VersionComboBox.SelectedIndex;
            switch (Version)
            {
                case FEVersion.Unknown:
                    MainChapterTextBox.ReadOnly = false;
                    ChapterDataTextBox.ReadOnly = false;
                    MainChapterTextBox.Enabled = true;
                    ChapterDataTextBox.Enabled = true;
                    ChapterCountSpinner.Enabled = true;
                    ChapterLengthSpinner.Enabled = true;
                    break;
                default:
                    MainChapterTextBox.ReadOnly = true;
                    ChapterDataTextBox.ReadOnly = true;
                    MainChapterTextBox.Enabled = false;
                    ChapterDataTextBox.Enabled = false;
                    ChapterCountSpinner.Enabled = false;
                    ChapterLengthSpinner.Enabled = false;

                    MainChapterTextBox.Text = string.Format("0x{0}", TilesetProcessing.DEFAULT_CHAPTER_DATA[Version].MainChapterTable.ToString("X8"));
                    ChapterDataTextBox.Text = string.Format("0x{0}", TilesetProcessing.DEFAULT_CHAPTER_DATA[Version].ChapterDataTable.ToString("X8"));
                    ChapterCountSpinner.Value = TilesetProcessing.DEFAULT_CHAPTER_DATA[Version].NumChapters;
                    ChapterLengthSpinner.Value = TilesetProcessing.DEFAULT_CHAPTER_DATA[Version].ChapterLength;
                    break;
            }
        }

        private void OpenFileButton_Click(object sender, EventArgs e)
        {
            if (!main_chapter_entry_valid)
            {
                MessageBox.Show("Main Chapter Table address is\nnot a valid hexadecimal number.", "Invalid Main Chapter Table", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            else if (!chapter_data_entry_valid)
            {
                MessageBox.Show("Chapter Data Table address is\nnot a valid hexadecimal number.", "Invalid Chapter Data Table", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            else if (LoadRomDialog.ShowDialog() == DialogResult.OK)
                load_game(LoadRomDialog.FileName);
        }

        void TilesetViewerForm_Shown(object sender, EventArgs e)
        {
            PaletteForm.Show();
            PaletteForm.Location = Point.Add(this.Location, new Size(this.Width / 2, this.Height / 2));
        }

        protected override void OnClosing(System.ComponentModel.CancelEventArgs e)
        {
            if (FERom != null)
            {
                FERom.CloseROM();
                FERom = null;
            }
            base.OnClosing(e);
        }

        void PaletteForm_PaletteChanged(object sender, EventArgs e)
        {
            if (tileset_exists && PaletteForm.palette != null && tileset_palette.Length == PaletteForm.palette.Length)
            {
                int index = PaletteListBox.SelectedIndex;
                tileset.set_palette(index, PaletteForm.palette);
                refresh_palette();
            }
        }

        void RawTilesetPictureBox_Paint(object sender, PaintEventArgs e)
        {
            if (RawTilesetImg != null)
            {
                e.Graphics.DrawImage(RawTilesetImg, new Rectangle(Point.Empty, RawTilesetImg.Size),
                    new Rectangle(Point.Empty, RawTilesetImg.Size), GraphicsUnit.Pixel);
                if (RawTilesetUnusedCheckBox.Checked)
                {
                    // Get a palette that's used for each subtile
                    var palettes = tileset.subtile_palettes();
                    if (palettes != null)
                        for (short i = 0; i < palettes.Length; i++)
                        {
                            int x = i % 32;
                            int y = i / 32;

                            Rectangle dest_rect = new Rectangle(new Point(x * 8, y * 8), new Size(8, 8));
                            if (palettes[i].IsNothing)
                                using (Brush b = new SolidBrush(Color.FromArgb(128, Color.Black)))
                                    e.Graphics.FillRectangle(b, dest_rect);
                        }
                }
            }
        }

        void TilesetPictureBox_Paint(object sender, PaintEventArgs e)
        {
            if (tileset_exists && this.tileset.Data.GraphicData != null && PaletteListBox.SelectedIndex >= 0 &&
                TilesetImg != null)
            {
                e.Graphics.DrawImage(TilesetImg, new Rectangle(Point.Empty, TilesetImg.Size),
                    new Rectangle(Point.Empty, TilesetImg.Size), GraphicsUnit.Pixel);
            }
        }
        #endregion

        /*private void get_fe7_terrain_stats_to_clipboard()
        {
            FEVersion version = FEVersion.FE7;

            GBAROM rom = new GBAROM();
            rom.OpenROM("FE7.gba");
            string stats = Terrain_Stats.parsed_terrain_stats(rom, version);
            rom.CloseROM();

            Clipboard.SetText(stats);
        }*/
    }
}
