﻿using System;
using System.Drawing;
using System.Drawing.Imaging;
using Nintenlord.Forms.Utility;

namespace GBAFE_Tilesets
{
    class FEGraphics
    {

        internal static unsafe void copy_indexed_pixels(Bitmap bmp, Bitmap src, int offset = 0)
        {
            int length2 = src.Width * src.Height;
            BitmapData srcData = src.LockBits(new Rectangle(new Point(), src.Size), ImageLockMode.WriteOnly, src.PixelFormat);
            byte* source = (byte*)srcData.Scan0;

            copy_indexed_pixels(bmp, source, length2, offset);

            src.UnlockBits(srcData);
        }
        static unsafe void copy_indexed_pixels(Bitmap bmp, byte* source, int length2, int offset)
        {
            int length1 = bmp.Width * bmp.Height;
            BitmapData bData = bmp.LockBits(new Rectangle(new Point(), bmp.Size), ImageLockMode.WriteOnly, bmp.PixelFormat);
            byte* bitmap = (byte*)bData.Scan0;

            copy_indexed_pixels(bitmap, length1, source, length2, offset);

            bmp.UnlockBits(bData);
        }
        internal static unsafe void copy_indexed_pixels(byte[] bmp, byte[] src, int offset = 0)
        {
            fixed (byte* p1 = bmp)
                fixed (byte* p2 = src)
                    copy_indexed_pixels(p1, bmp.Length, p2, src.Length, offset);
        }
        static unsafe void copy_indexed_pixels(byte* bitmap, int length1, byte* source, int length2, int offset)
        {
            for (int i = 0; i + offset < length1 && i < length2; i++)
                bitmap[i + offset] = source[i];
        }

        /// <summary>
        /// Copies an 8x8 subtile of a tileset image to an 8x8 indexed bitmap.
        /// </summary>
        internal static unsafe void copy_indexed_subtile(Bitmap bmp, Bitmap src, short image_src, byte palette_index, bool h_flip, bool v_flip, Color[] tileset_palette)
        {
            int length2 = src.Width * src.Height;
            BitmapData srcData = src.LockBits(new Rectangle(new Point(), src.Size), ImageLockMode.WriteOnly, src.PixelFormat);
            byte* source = (byte*)srcData.Scan0;

            copy_indexed_subtile(bmp, source, length2, image_src, palette_index, h_flip, v_flip, tileset_palette);

            src.UnlockBits(srcData);
        }
        /// <summary>
        /// Copies an 8x8 subtile of a tileset byte array to an 8x8 indexed bitmap.
        /// </summary>
        static unsafe void copy_indexed_subtile(Bitmap bmp, byte* source, int length2, short image_src, byte palette_index, bool h_flip, bool v_flip, Color[] tileset_palette)
        {
            // Copies palette from the tileset palette list
            Color[] palette = new Color[16];
            if (16 * palette_index >= 256)
            {
                for (int i = 0; i < 16; i++)
                    palette[i] = Color.Magenta;
            }
            else
                Array.Copy(tileset_palette, 16 * palette_index, palette, 0, 16);
            bmp.InsertColorsToPalette(palette);

            // Copies the tile graphic from the tileset to the tile
            int height = bmp.Height;
            int width = bmp.Width;
            BitmapData bData = bmp.LockBits(new Rectangle(new Point(), bmp.Size), ImageLockMode.WriteOnly, bmp.PixelFormat);
            byte* bitmap = (byte*)bData.Scan0;
            // If the tileset is the wrong size for some reason, or the src tile is out of bounds, do this instead of crashing
            if (length2 != 256 * 256 || image_src >= 1024 || width * height != 8 * 8)
            {
                for (int i = 0; i < width * height; i++)
                    bitmap[i] = (byte)(i % 16);
            }
            else
            {
                int ox = image_src % 32 * 8;
                int oy = image_src / 32 * 8;

                for (int y = 0; y < height; y++)
                    for (int x = 0; x < width; x++)
                    {
                        int src_offset = ((v_flip ? 7 - y : y) + oy) * 256 + ((h_flip ? 7 - x : x) + ox);
                        bitmap[y * 8 + x] = source[src_offset];
                    }
            }
            bmp.UnlockBits(bData);
        }

        /// <summary>
        /// Copies an 8x8 subtile of a tileset image to an indexed bitmap.
        /// </summary>
        internal static unsafe void copy_indexed_subtile(Bitmap bmp, Point loc, Bitmap src, short image_src, byte palette_index, bool h_flip, bool v_flip, Color[] tileset_palette)
        {
            int length2 = src.Width * src.Height;
            BitmapData srcData = src.LockBits(new Rectangle(new Point(), src.Size), ImageLockMode.WriteOnly, src.PixelFormat);
            byte* source = (byte*)srcData.Scan0;

            copy_indexed_subtile(bmp, loc, source, length2, image_src, palette_index, h_flip, v_flip, tileset_palette);

            src.UnlockBits(srcData);
        }
        /// <summary>
        /// Copies an 8x8 subtile of a tileset byte array to an indexed bitmap.
        /// </summary>
        static unsafe void copy_indexed_subtile(Bitmap bmp, Point loc, byte* source, int length2, short image_src, byte palette_index, bool h_flip, bool v_flip, Color[] tileset_palette)
        {
            // Copies the tile graphic from the tileset to the tile
            int height = bmp.Height;
            int width = bmp.Width;
            BitmapData bData = bmp.LockBits(new Rectangle(new Point(), bmp.Size), ImageLockMode.WriteOnly, bmp.PixelFormat);
            byte* bitmap = (byte*)bData.Scan0;
            // If the palette index is out of bounds, uses the last palette index if it doesn't exist; otherwise sets all pixels to index 0
            bool index_out_of_bounds = false;
            if (16 * palette_index >= 256)
            {
                if (tileset_palette.Length > 256 - 16)
                    index_out_of_bounds = true;
                else
                    palette_index = 15;
            }
            int ox = image_src % 32 * 8;
            int oy = image_src / 32 * 8;
            if (loc.X + 7 > width || (loc.Y + 7) > height || ((oy + 7) * 256 + (ox + 7)) >= length2)
            {
               // Haha mistakes
            }
            else if (index_out_of_bounds)
            {
                for (int y = 0; y < 8; y++)
                    for (int x = 0; x < 8; x++)
                    {
                        int bmp_offset = (y + loc.Y) * width + (x + loc.X);
                        bitmap[bmp_offset] = 0;
                    }
            }
            // If the tileset is the wrong size for some reason, or the src tile is out of bounds, do this instead of crashing
            else if (length2 != 256 * 256 || image_src >= 1024)
            {
                for (int y = 0; y < 8; y++)
                    for (int x = 0; x < 8; x++)
                    {
                        int bmp_offset = (y + loc.Y) * width + (x + loc.X);
                        bitmap[bmp_offset] = (byte)(bmp_offset % 16);
                    }
            }
            else
            {
                for (int y = 0; y < 8; y++)
                    for (int x = 0; x < 8; x++)
                    {
                        int bmp_offset = (y + loc.Y) * width + (x + loc.X);
                        int src_offset = ((v_flip ? 7 - y : y) + oy) * 256 + ((h_flip ? 7 - x : x) + ox);
                        bitmap[bmp_offset] = (byte)(source[src_offset] + palette_index * 16);
                    }
            }
            bmp.UnlockBits(bData);
        }
    }
}
