﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Imaging;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Nintenlord.ROMHacking.GBA;
using Nintenlord.ROMHacking.GBA.Graphics;
using Nintenlord.Forms.Utility;

namespace GBAFE_Tilesets
{
    static class TilesetProcessing
    {
        internal readonly static Dictionary<FEVersion, FEChapterFormatInfo> DEFAULT_CHAPTER_DATA = new Dictionary<FEVersion, FEChapterFormatInfo>
        {
            { FEVersion.FE6, new FEChapterFormatInfo(0x664398, 0x6637A4,  68, 45) },
            { FEVersion.FE7, new FEChapterFormatInfo(0xC9C9C8, 0xC9A200, 152, 67) },
            { FEVersion.FE8, new FEChapterFormatInfo(0x8B363C, 0x8B0890, 148, 79) }
        };
        readonly static Dictionary<FEVersion, int> CHAPTER_DATA_TILESET_OFFSET = new Dictionary<FEVersion, int>
        {
            { FEVersion.FE6, 4 },
            { FEVersion.FE7, 4 },
            { FEVersion.FE8, 4 },
            { FEVersion.Unknown, 4 },
        };

        internal static HashSet<Tuple<byte, byte, byte, byte>> get_unique_tilesets(GBAROM rom, FEVersion version, FEChapterFormatInfo chapter_info)
        {
            if (rom.Length < chapter_info.ChapterDataTable || rom.Length < chapter_info.MainChapterTable)
                return null;
            int entries = chapter_info.NumChapters;
            HashSet<Tuple<byte, byte, byte, byte>> unique_tilesets = new HashSet<Tuple<byte, byte, byte, byte>>();
            for(int i = 0; i < entries; i++)
            {
                byte[] chapter_data = rom.GetData(chapter_info.ChapterDataTable + i * chapter_info.ChapterLength, chapter_info.ChapterLength);
                int offset = CHAPTER_DATA_TILESET_OFFSET[version];

                var graphics_pointer1 = rom.ReadPointer(chapter_info.MainChapterTable + chapter_data[offset + 0] * 4);
                var graphics_pointer2 = rom.ReadPointer(chapter_info.MainChapterTable + chapter_data[offset + 1] * 4);
                var palette_pointer = rom.ReadPointer(chapter_info.MainChapterTable + chapter_data[offset + 2] * 4);
                var data_pointer = rom.ReadPointer(chapter_info.MainChapterTable + chapter_data[offset + 3] * 4);
                if (graphics_pointer1 < rom.Length && graphics_pointer2 < rom.Length &&
                        (rom.DecompressLZ77CompressedData(graphics_pointer1) != null || rom.DecompressLZ77CompressedData(graphics_pointer2) != null) &&
                        data_pointer < rom.Length && palette_pointer + 320 <= rom.Length && rom.DecompressLZ77CompressedData(data_pointer) != null)
                    unique_tilesets.Add(new Tuple<byte, byte, byte, byte>(
                        chapter_data[offset + 0],
                        chapter_data[offset + 1],
                        chapter_data[offset + 2],
                        chapter_data[offset + 3]));
            }
            return unique_tilesets;
        }

        internal static byte[] load_tileset_data(GBAROM rom, FEChapterFormatInfo chapter_info, TilesetCoreData tileset)
        {
            return load_tileset_data(rom, chapter_info, tileset.DataOffsets);
        }
        internal static byte[] load_tileset_data(GBAROM rom, FEChapterFormatInfo chapter_info, Tuple<byte, byte, byte> tuple)
        {
            int data_offset = tuple.Item3;
            var pointer = rom.ReadPointer(chapter_info.MainChapterTable + data_offset * 4);
            return rom.DecompressLZ77CompressedData(pointer);
        }

        internal static byte[] load_tileset_palette(GBAROM rom, FEChapterFormatInfo chapter_info, Tuple<byte, byte, byte, byte> tuple)
        {
            return load_tileset_palette(rom, chapter_info, tuple.Item3);
        }
        internal static byte[] load_tileset_palette(GBAROM rom, FEChapterFormatInfo chapter_info, byte pal_offset)
        {
            var pointer = rom.ReadPointer(chapter_info.MainChapterTable + pal_offset * 4);
            return rom.GetData(pointer, 320);
        }

        internal static byte[] load_tileset_graphic_data(GBAROM rom, FEChapterFormatInfo chapter_info, TilesetCoreData tileset)
        {
            return load_tileset_graphic_data(rom, chapter_info, tileset.DataOffsets);
        }
        internal static byte[] load_tileset_graphic_data(GBAROM rom, FEChapterFormatInfo chapter_info, Tuple<byte, byte, byte> tuple)
        {
            // Load first graphic object; if it doesn't exist, return null
            byte[] raw_graphic = load_tileset_object(rom, chapter_info, tuple.Item1);
            if (raw_graphic == null)
                return null;

            byte[] raw_bmp = new byte[(256 * 256) / 2];
            FEGraphics.copy_indexed_pixels(raw_bmp, raw_graphic);
            // Load second graphic object, if it exists
            raw_graphic = load_tileset_object(rom, chapter_info, tuple.Item2);
            if (raw_graphic != null)
            {
                FEGraphics.copy_indexed_pixels(raw_bmp, raw_graphic, (256 * 128) / 2);
            }
            return raw_bmp;
        }
        internal static Bitmap load_raw_tileset_graphic(GBAROM rom, FEChapterFormatInfo chapter_info, TilesetData tileset, Color[] tileset_palette)
        {
            byte[] raw_graphic = load_tileset_graphic_data(rom, chapter_info, tileset.Data.DataOffsets);
            if (raw_graphic == null)
                return null;
            int emptyGraphicsBlocks;

            Bitmap bmp = new Bitmap(256, 256, PixelFormat.Format8bppIndexed);
            // Get a palette that's used for each subtile
            var palettes = tileset.subtile_palettes();
            if (palettes == null)
                return null;
            // Copy subtiles
            using (Bitmap src = GBAGraphics.ToBitmap(raw_graphic, raw_graphic.Length, 0, tileset_palette, 32 * 8,
                GraphicsMode.Tile4bit, out emptyGraphicsBlocks))
            {
                for (short i = 0; i < palettes.Length; i++)
                {
                    int x = i % 32;
                    int y = i / 32;

                    byte palette_index = palettes[i].IsSomething && palettes[i].Value >= 5 ? (byte)255 : (byte)(palettes[i].IsNothing ? 0 : palettes[i].Value);
                    FEGraphics.copy_indexed_subtile(bmp, new Point(x * 8, y * 8), src, i, palette_index, false, false, tileset_palette);
                }
            }
            return bmp;
        }
        internal static Bitmap load_tileset_graphic(GBAROM rom, FEChapterFormatInfo chapter_info, TilesetData tileset, Color[] tileset_palette)
        {
            byte[] raw_graphic = load_tileset_graphic_data(rom, chapter_info, tileset.Data.DataOffsets);
            if (raw_graphic == null)
                return null;
            int emptyGraphicsBlocks;

            if (tileset.Data.Tsa.Length / 9 > 1024)
                return null;
            Bitmap bmp = new Bitmap(512, 512, PixelFormat.Format8bppIndexed);
            // Copy subtiles
            using (Bitmap src = GBAGraphics.ToBitmap(raw_graphic, raw_graphic.Length, 0, tileset_palette, 32 * 8,
                GraphicsMode.Tile4bit, out emptyGraphicsBlocks))
            {
                for (int i = 0; i < tileset.Data.Tsa.Length / 9; i++)
                {
                    int x = i % 32;
                    int y = i / 32;

                    for (int j = 0; j < 4; j++)
                    {
                        ushort data = (ushort)((tileset.Data.Tsa[i * 8 + j * 2 + 1] << 8) + tileset.Data.Tsa[i * 8 + j * 2 + 0]);
                        byte palette_index = (byte)((data & 0xF000) >> 12); // First 4
                        bool v_flip = ((data & 0x0800) >> 11) == 1; // Bit 5
                        bool h_flip = ((data & 0x0400) >> 10) == 1; // Bit 6
                        short image_src = (short)(data & 0x03FF); // Last 10

                        palette_index = (byte)(palette_index >= 5 ? 255 : palette_index);
                        FEGraphics.copy_indexed_subtile(bmp, new Point(x * 16 + (j % 2) * 8, y * 16 + (j / 2) * 8),
                            src, image_src, palette_index, h_flip, v_flip, tileset_palette);
                    }
                }
            }
            return bmp;
        }
        static byte[] load_tileset_object(GBAROM rom, FEChapterFormatInfo chapter_info, byte offset)
        {
            var pointer = rom.ReadPointer(chapter_info.MainChapterTable + offset * 4);
            return rom.DecompressLZ77CompressedData(pointer);
        }

        internal static void set_tileset_palette(Bitmap bmp, Color[] tileset_palette, bool fog = false)
        {
            Color[] actual_palette = new Color[tileset_palette.Length];
            if (fog)
                Array.Copy(tileset_palette, 16 * 5, actual_palette, 0, tileset_palette.Length - 16 * 5);
            else
                Array.Copy(tileset_palette, actual_palette, tileset_palette.Length);

            Color[] palette = new Color[256];
            if (actual_palette.Length > 256 - 16)
                Array.Copy(actual_palette, palette, Math.Min(actual_palette.Length, palette.Length));
            else
            {
                Array.Copy(actual_palette, palette, actual_palette.Length);
                for (int i = 256 - 16; i < 256; i++)
                    palette[i] = Color.Magenta;
            }
            bmp.InsertColorsToPalette(palette);
        }
    }
}
