﻿using System;
using System.Drawing;
using System.Linq;

namespace GBAFE_Tilesets
{
    public enum Color_Bit_Depth : int
    {
        Red1 = 0,
        Red5 = 1 << 2,
        Red8 = 1 << 2 | 1 << 1 | 1 << 0,
        Red16 = 1 << 3 | 1 << 2 | 1 << 1 | 1 << 0,
        Green1 = 0,
        Green5 = 1 << 6,
        Green6 = 1 << 6 | 1 << 4,
        Green8 = 1 << 6 | 1 << 5 | 1 << 4,
        Green16 = 1 << 7 | 1 << 6 | 1 << 5 | 1 << 4,
        Blue1 = 0,
        Blue5 = 1 << 10,
        Blue8 = 1 << 10 | 1 << 9 | 1 << 8,
        Blue16 = 1 << 11 | 1 << 10 | 1 << 9 | 1 << 8,

        R5G5B5 = Red5 | Green5 | Blue5,
        R5G6B5 = Red5 | Green6 | Blue5,
        R8G8B8 = Red8 | Green8 | Blue8,

        None = 1 << 12,
    }
    public enum Color_Channels { Red, Green, Blue, Alpha }
    static class YetiGraphics
    {
        public static Color color_from_hsl(double hue, double saturation, double lightness, Color_Bit_Depth bit_depth = Color_Bit_Depth.R8G8B8)
        {
            double r, g, b;
            double h = hue % 360;
            if (h < 0)
                h += 360;

            if (lightness <= 0)
                r = g = b = 0;
            else if (saturation <= 0)
                r = g = b = lightness;
            else
            {
                double hf = hue / 60.0;
                int i = (int)Math.Floor(hf);
                double chroma = (1 - Math.Abs(lightness * 2 - 1)) * saturation;
                double x = chroma * (1 - Math.Abs(hf % 2 - 1));
                double m = lightness - (chroma / 2);
                switch (i)
                {

                    // Red dominant: Red to yellow
                    case 0:
                    case 6:
                        r = chroma + m;
                        g = x + m;
                        b = m;
                        break;
                    // Green dominant: Yellow to green
                    case 1:
                        r = x + m;
                        g = chroma + m;
                        b = m;
                        break;
                    // Green dominant: Green to cyan
                    case 2:
                        r = m;
                        g = chroma + m;
                        b = x + m;
                        break;
                    // Blue dominant: Cyan to blue
                    case 3:
                        r = m;
                        g = x + m;
                        b = chroma + m;
                        break;
                    // Blue dominant: Blue to Magenta
                    case 4:
                        r = x + m;
                        g = m;
                        b = chroma + m;
                        break;
                    // Red dominant: Magenta to red
                    case 5:
                    case -1:
                        r = chroma + m;
                        g = m;
                        b = x + m;
                        break;
                    // Something has gone wrong
                    default:
                        r = g = b = lightness;
                        break;
                }
            }

            return bit_depth_fixed_color(r, g, b, bit_depth);
        }
        public static Color color_from_hsv(double hue, double saturation, double value, Color_Bit_Depth bit_depth = Color_Bit_Depth.R8G8B8)
        {
            double r, g, b;
            double h = hue % 360;
            if (h < 0)
                h += 360;

            if (value <= 0)
                r = g = b = 0;
            else if (saturation <= 0)
                r = g = b = value;
            else
            {
                double hf = hue / 60.0;
                int i = (int)Math.Floor(hf);
                double f = hf - i;
                double pv = value * (1 - saturation);
                double qv = value * (1 - saturation * f);
                double tv = value * (1 - saturation * (1 - f));
                switch (i)
                {

                    // Red dominant: Red to yellow
                    case 0:
                    case 6:
                        r = value;
                        g = tv;
                        b = pv;
                        break;
                    // Green dominant: Yellow to green
                    case 1:
                        r = qv;
                        g = value;
                        b = pv;
                        break;
                    // Green dominant: Green to cyan
                    case 2:
                        r = pv;
                        g = value;
                        b = tv;
                        break;
                    // Blue dominant: Cyan to blue
                    case 3:
                        r = pv;
                        g = qv;
                        b = value;
                        break;
                    // Blue dominant: Blue to Magenta
                    case 4:
                        r = tv;
                        g = pv;
                        b = value;
                        break;
                    // Red dominant: Magenta to red
                    case 5:
                    case -1:
                        r = value;
                        g = pv;
                        b = qv;
                        break;
                    // Something has gone wrong
                    default:
                        r = g = b = value;
                        break;
                }
            }

            return bit_depth_fixed_color(r, g, b, bit_depth);
        }

        private static Color bit_depth_fixed_color(double r, double g, double b, Color_Bit_Depth bit_depth)
        {
            int r_depth = channel_bit_depth(bit_depth, Color_Channels.Red);
            int g_depth = channel_bit_depth(bit_depth, Color_Channels.Green);
            int b_depth = channel_bit_depth(bit_depth, Color_Channels.Blue);

            return Color.FromArgb(
                Math.Max(0, Math.Min(255, (int)(r * (r_depth - 1)) * (256 / r_depth))),
                Math.Max(0, Math.Min(255, (int)(g * (g_depth - 1)) * (256 / g_depth))),
                Math.Max(0, Math.Min(255, (int)(b * (b_depth - 1)) * (256 / b_depth))));
        }

        public static Color convert_color_bit_depth(Color color, Color_Bit_Depth new_bit_depth, Color_Bit_Depth old_bit_depth)
        {
            int new_r_depth = channel_bit_depth(new_bit_depth, Color_Channels.Red);
            int new_g_depth = channel_bit_depth(new_bit_depth, Color_Channels.Green);
            int new_b_depth = channel_bit_depth(new_bit_depth, Color_Channels.Blue);

            int old_r_depth = channel_bit_depth(old_bit_depth, Color_Channels.Red);
            int old_g_depth = channel_bit_depth(old_bit_depth, Color_Channels.Green);
            int old_b_depth = channel_bit_depth(old_bit_depth, Color_Channels.Blue);

            return Color.FromArgb(
                (color.R / (256 / new_r_depth)) * (256 / new_r_depth),
                (color.G / (256 / new_g_depth)) * (256 / new_g_depth),
                (color.B / (256 / new_b_depth)) * (256 / new_b_depth));
        }
        public static int channel_bit_depth(Color_Bit_Depth bit_depth, Color_Channels channel)
        {
            switch (channel)
            {
                case Color_Channels.Red:
                    return 1 << ((int)(bit_depth & Color_Bit_Depth.Red16) + 1);
                case Color_Channels.Green:
                    return 1 << (((int)(bit_depth & Color_Bit_Depth.Green16) >> 4) + 1);
                case Color_Channels.Blue:
                    return 1 << (((int)(bit_depth & Color_Bit_Depth.Blue16) >> 8) + 1);
            }
            return 256;
        }

        internal static byte[] color_array_to_hex(Color[] palette)
        {
            return palette.SelectMany(x =>
                {
                    int hex = ((x.B >> 3) << 10) + ((x.G >> 3) << 5) + (x.R >> 3);
                    return new byte[] { (byte)(hex & 0xFF), (byte)((hex & 0xFF00) >> 8) };
                }).ToArray();
        }
        internal static Color[] hex_to_color_array(byte[] palette)
        {
            return Enumerable.Range(0, palette.Length / 2).Select(x =>
                {
                    int hex = (palette[x * 2 + 1] << 8) + palette[x * 2];
                    return Color.FromArgb(255, (hex & 0x001F) << 3, (hex & 0x03E0) >> 2, (hex & 0x7C00) >> 7);
                }).ToArray();
        }
    }
}
